import threading
import socket
import errno
import ssl
import time
import queue

from stomp.const import EOL
from stomp.const import ERROR
from stomp.const import RECEIPT
from stomp.const import CONNECTED
from stomp.const import DISCONNECT
from stomp.const import DISCONNECTED
from stomp.utils import parse_raw_frames
from stomp.utils import parse_receipt_id
from stomp.streambuffer import StreamBuffer
from stomp.exceptions import NoReceiptGiven
from stomp.exceptions import AuthenticationError
from stomp.exceptions import StompException
from stomp.exceptions import Disconnected
from stomp.datastructures import RawFrame
from stomp.frame import Connect as ConnectFrame
from stomp.frame import Frame


class Connection(object):
    """Represents a connection to the STOMP server."""
    default_chunk_size = 1024**2
    default_timeout = 10000

    def __init__(self, host, port, chunk_size=None, enable_ssl=False,
        encoding='utf-8'):
        self._ssl = None
        if enable_ssl:
            self._ssl = ssl.create_default_context(
            ssl.Purpose.CLIENT_AUTH)
        self._chunk_size = chunk_size or self.default_chunk_size
        self._last_heartbeat = None
        self._encoding = 'utf-8'
        self._host = host
        self._port = port

        # Threading events
        self.__queue = queue.Queue()
        self.__disconnected_event = threading.Event()
        self.__lock = threading.Lock()
        self.__recv_thread = threading.Thread(target=self.__main_event_loop)
        self.__recv_thread.daemon = True

        # Setup the buffer.
        self._buffer = StreamBuffer(frame_factory=Frame.parse,
            frame_queue=self.__queue)

    def connect(self):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.connect((self._host, self._port))
        if self._ssl is not None:
            self._socket = self._ssl.wrap_socket(self._socket)

    def login(self, virtualhost, username, password, heartbeats=None):
        """Login to the STOMP server.

        Args:
            virtualhost (str): the virtual host to connect to.
            username (str): the user to connect as.
            password (str): authenticates the user.
            heartbeats (tuple): specifies the desired heartbeat protocol.

        Returns:
            None

        Raises:
            stomp.exceptions.AuthenticationError: invalid credentials.
        """
        if heartbeats is None:
            heartbeats = (0,0)
        frame = ConnectFrame(virtualhost, username, password,
            server_heartbeats=heartbeats[1],
            client_heartbeats=heartbeats[0])
        self._socket.send(bytes(frame))
        self._socket.setblocking(1)

        self._buffer.write(self._socket.recv(1024))
        frame = self._buffer.read()
        if frame.frame_type == ERROR:
            msg = frame.headers.get('message') or 'No message supplied'
            raise AuthenticationError(msg)

        # Assume CONNECT is succesful, and set unset the disconnected
        # event, so the main event loop won't exit immediatly.
        self.__disconnected_event.clear()

        # Now we set the socket to non-blocking to start the main
        # loop of receiving data.
        self._socket.setblocking(0)


    def start(self):
        # Start the main event loop.
        self.__recv_thread.start()

    def send(self, raw_frame):
        """Sends a frame to the STOMP server.

        Args:
            frame (bytes): a byte-sequence holding a STOMP frame.

        Returns:
            None

        Raises:
            stomp.exceptions.Disconnected: not connected to STOMP server.
            stomp.exceptions.StompException: server returned an ``ERROR``
                frame.
        """
        if not self.is_connected():
            raise Disconnected("Not connected to STOMP server.")
        with self.__lock:
            try:
                self._socket.send(raw_frame)
            except Exception as e:
                if getattr(e, 'errno', None) == errno.ECONNRESET:
                    self.__disconnected_event.set()
                    raise Disconnected("Not connected to STOMP server.")
                raise

    def is_connected(self):
        """Return a boolean indicating if :class:`Connection` is currently
        connected to the STOMP server.
        """
        return not self.__disconnected_event.is_set()

    def on_error(self, frame):
        """Performs cleanup after the STOMP server sent us an ``ERROR``
        frame.
        """
        # The STOMP protocol mandates that the connection is closed after
        # an ERROR frame.
        self.__disconnected_event.set()

    def __main_event_loop(self):
        while True:
            if not self.is_connected():
                break
            try:
                self.__main_event()
            except Exception as e:
                if getattr(e, 'errno', None) == errno.ECONNRESET:
                    break
                if not getattr(e, 'errno', None) == errno.EAGAIN:
                    raise
                time.sleep(0.01)

    def __main_event(self):
        self._process_data(self._recv())

    def _recv(self):
        with self.__lock:
            try:
                data = self._socket.recv(self._chunk_size)
            except Exception as e:
                if getattr(e, 'errno', None) == errno.ECONNRESET:
                    self.__disconnected_event.set()
                raise
            self._last_heartbeat = time.time()
        return data.decode(encoding=self._encoding)

    def _process_data(self, stream):
        if stream == EOL or (stream.count(EOL) == len(stream)):
            return
        self._buffer.write(stream)
        try:
            frame = self._buffer.read(blocking=False)
        except Exception as e:
            if not getattr(e, 'errno', None) == errno.EAGAIN:
                raise
            return

        if frame.frame_type in (ERROR, DISCONNECT):
            self.on_error(frame)
        self.__queue.put(frame)
